@@ -107,11 +107,19 @@ class TestAclOps(OvnNorthboundTest):
     def setUp(self):
         super(TestAclOps, self).setUp()
         self.switch = self.useFixture(fixtures.LogicalSwitchFixture()).obj
+        self.port_group = self.useFixture(fixtures.PortGroupFixture()).obj
+
+    def _acl_add(self, entity, *args, **kwargs):
+        self.assertIn(entity, ['lswitch', 'port_group'])
+        if entity == 'lswitch':
+            cmd = self.api.acl_add(self.switch.uuid, *args, **kwargs)
+            resource = self.switch
+        else:
+            cmd = self.api.pg_acl_add(self.port_group.uuid, *args, **kwargs)
+            resource = self.port_group
 
-    def _acl_add(self, *args, **kwargs):
-        cmd = self.api.acl_add(self.switch.uuid, *args, **kwargs)
         aclrow = cmd.execute(check_error=True)
-        self.assertIn(aclrow._row, self.switch.acls)
+        self.assertIn(aclrow._row, resource.acls)
         self.assertEqual(cmd.direction, aclrow.direction)
         self.assertEqual(cmd.priority, aclrow.priority)
         self.assertEqual(cmd.match, aclrow.match)
@@ -119,43 +127,50 @@ class TestAclOps(OvnNorthboundTest):
         return aclrow
 
     def test_acl_add(self):
-        self._acl_add('from-lport', 0, 'output == "fake_port" && ip',
-                      'drop')
+        self._acl_add('lswitch', 'from-lport', 0,
+                      'output == "fake_port" && ip', 'drop')
 
     def test_acl_add_exists(self):
-        args = ('from-lport', 0, 'output == "fake_port" && ip', 'drop')
+        args = ('lswitch', 'from-lport', 0, 'output == "fake_port" && ip',
+                'drop')
         self._acl_add(*args)
         self.assertRaises(RuntimeError, self._acl_add, *args)
 
     def test_acl_add_may_exist(self):
         args = ('from-lport', 0, 'output == "fake_port" && ip', 'drop')
-        row = self._acl_add(*args)
-        row2 = self._acl_add(*args, may_exist=True)
+        row = self._acl_add('lswitch', *args)
+        row2 = self._acl_add('lswitch', *args, may_exist=True)
         self.assertEqual(row, row2)
 
     def test_acl_add_extids(self):
         external_ids = {'mykey': 'myvalue', 'yourkey': 'yourvalue'}
-        acl = self._acl_add('from-lport', 0, 'output == "fake_port" && ip',
+        acl = self._acl_add('lswitch',
+                            'from-lport', 0, 'output == "fake_port" && ip',
                             'drop', **external_ids)
         self.assertEqual(external_ids, acl.external_ids)
 
     def test_acl_del_all(self):
-        r1 = self._acl_add('from-lport', 0, 'output == "fake_port"', 'drop')
+        r1 = self._acl_add('lswitch', 'from-lport', 0, 'output == "fake_port"',
+                           'drop')
         self.api.acl_del(self.switch.uuid).execute(check_error=True)
         self.assertNotIn(r1.uuid, self.api.tables['ACL'].rows)
         self.assertEqual([], self.switch.acls)
 
     def test_acl_del_direction(self):
-        r1 = self._acl_add('from-lport', 0, 'output == "fake_port"', 'drop')
-        r2 = self._acl_add('to-lport', 0, 'output == "fake_port"', 'allow')
+        r1 = self._acl_add('lswitch', 'from-lport', 0,
+                           'output == "fake_port"', 'drop')
+        r2 = self._acl_add('lswitch', 'to-lport', 0,
+                           'output == "fake_port"', 'allow')
         self.api.acl_del(self.switch.uuid, 'from-lport').execute(
             check_error=True)
         self.assertNotIn(r1, self.switch.acls)
         self.assertIn(r2, self.switch.acls)
 
     def test_acl_del_direction_priority_match(self):
-        r1 = self._acl_add('from-lport', 0, 'output == "fake_port"', 'drop')
-        r2 = self._acl_add('from-lport', 1, 'output == "fake_port"', 'allow')
+        r1 = self._acl_add('lswitch', 'from-lport', 0,
+                           'output == "fake_port"', 'drop')
+        r2 = self._acl_add('lswitch', 'from-lport', 1,
+                           'output == "fake_port"', 'allow')
         cmd = self.api.acl_del(self.switch.uuid,
                                'from-lport', 0, 'output == "fake_port"')
         cmd.execute(check_error=True)
@@ -171,12 +186,35 @@ class TestAclOps(OvnNorthboundTest):
                           priority=0)
 
     def test_acl_list(self):
-        r1 = self._acl_add('from-lport', 0, 'output == "fake_port"', 'drop')
-        r2 = self._acl_add('from-lport', 1, 'output == "fake_port2"', 'allow')
+        r1 = self._acl_add('lswitch', 'from-lport', 0,
+                           'output == "fake_port"', 'drop')
+        r2 = self._acl_add('lswitch', 'from-lport', 1,
+                           'output == "fake_port2"', 'allow')
         acls = self.api.acl_list(self.switch.uuid).execute(check_error=True)
         self.assertIn(r1, acls)
         self.assertIn(r2, acls)
 
+    def test_pg_acl_add(self):
+        self._acl_add('port_group', 'from-lport', 0,
+                      'output == "fake_port" && ip', 'drop')
+
+    def test_pg_acl_del_all(self):
+        r1 = self._acl_add('port_group', 'from-lport', 0,
+                           'output == "fake_port"', 'drop')
+        self.api.pg_acl_del(self.port_group.uuid).execute(check_error=True)
+        self.assertNotIn(r1.uuid, self.api.tables['ACL'].rows)
+        self.assertEqual([], self.port_group.acls)
+
+    def test_pg_acl_list(self):
+        r1 = self._acl_add('port_group', 'from-lport', 0,
+                           'output == "fake_port"', 'drop')
+        r2 = self._acl_add('port_group', 'from-lport', 1,
+                           'output == "fake_port2"', 'allow')
+        acls = self.api.pg_acl_list(self.port_group.uuid).execute(
+            check_error=True)
+        self.assertIn(r1, acls)
+        self.assertIn(r2, acls)
+
 
 class TestLspOps(OvnNorthboundTest):
     def setUp(self):
@@ -342,7 +380,7 @@ class TestLspOps(OvnNorthboundTest):
 
     def test_lsp_set_addresses(self):
         lsp = self._lsp_add(self.switch, None)
-        for addr in ('dynamic', 'unknown', 'router',
+        for addr in ('dynamic', 'unknown', 'router', 'de:ad:be:ef:4d:ad',
                      'de:ad:be:ef:4d:ad 192.0.2.1'):
             self.api.lsp_set_addresses(lsp.name, [addr]).execute(
                 check_error=True)
@@ -351,7 +389,7 @@ class TestLspOps(OvnNorthboundTest):
     def test_lsp_set_addresses_invalid(self):
         self.assertRaises(
             TypeError,
-            self.api.lsp_set_addresses, 'fake', '01:02:03:04:05:06')
+            self.api.lsp_set_addresses, 'fake', 'invalidaddress')
 
     def test_lsp_get_addresses(self):
         addresses = [
@@ -1087,3 +1125,183 @@ class TestCommonDbOps(OvnNorthboundTest):
         # should be a NoOp, not fail
         self.api.db_remove('Logical_Switch', self.switch.uuid, 'ports',
                            "badvalue").execute(check_error=True)
+
+
+class TestDnsOps(OvnNorthboundTest):
+    def _dns_add(self, *args, **kwargs):
+        dns = self.useFixture(fixtures.DnsFixture(*args, **kwargs)).obj
+        return dns
+
+    def test_dns_get(self):
+        dns = self._dns_add()
+        found = self.api.dns_get(dns.uuid).execute(
+            check_error=True)
+        self.assertEqual(dns, found)
+
+    def test_dns_get_no_exist(self):
+        cmd = self.api.dns_get("noexist")
+        self.assertRaises(idlutils.RowNotFound, cmd.execute, check_error=True)
+
+    def test_dns_add(self):
+        self._dns_add()
+
+    def test_dns_add_ext_ids(self):
+        ext_ids = {'net-id': '1', 'other-id': '2'}
+        dns = self._dns_add(external_ids=ext_ids)
+        self.assertEqual(ext_ids, dns.external_ids)
+
+    def test_dns_list(self):
+        dnses = {self._dns_add() for d in range(3)}
+        dnses_set = set(
+            self.api.dns_list().execute(check_error=True))
+        self.assertTrue(dnses.issubset(dnses_set))
+
+    def test_dns_set_records(self):
+        dns = self._dns_add()
+        records = {'a': 'one', 'b': 'two'}
+        self.api.dns_set_records(
+            dns.uuid, **records).execute(check_error=True)
+        dns = self.api.dns_get(dns.uuid).execute(
+            check_error=True)
+        self.assertEqual(records, dns.records)
+        self.api.dns_set_records(
+            dns.uuid, **{}).execute(check_error=True)
+        self.assertEqual({}, dns.records)
+
+    def test_dns_set_external_ids(self):
+        dns = self._dns_add()
+        external_ids = {'a': 'one', 'b': 'two'}
+        self.api.dns_set_external_ids(
+            dns.uuid, **external_ids).execute(check_error=True)
+        dns = self.api.dns_get(dns.uuid).execute(
+            check_error=True)
+        self.assertEqual(external_ids, dns.external_ids)
+        self.api.dns_set_external_ids(
+            dns.uuid, **{}).execute(check_error=True)
+        self.assertEqual({}, dns.external_ids)
+
+    def test_dns_add_remove_records(self):
+        dns = self._dns_add()
+        self.api.dns_add_record(dns.uuid, 'a', 'one').execute()
+        self.api.dns_add_record(dns.uuid, 'b', 'two').execute()
+        dns = self.api.dns_get(dns.uuid).execute(
+            check_error=True)
+        records = {'a': 'one', 'b': 'two'}
+        self.assertEqual(records, dns.records)
+        self.api.dns_remove_record(dns.uuid, 'a').execute()
+        records.pop('a')
+        self.assertEqual(records, dns.records)
+        self.api.dns_remove_record(dns.uuid, 'b').execute()
+        self.assertEqual({}, dns.records)
+
+
+class TestLsDnsOps(OvnNorthboundTest):
+    def _dns_add(self, *args, **kwargs):
+        dns = self.useFixture(fixtures.DnsFixture(*args, **kwargs)).obj
+        return dns
+
+    def _ls_add(self, *args, **kwargs):
+        fix = self.useFixture(fixtures.LogicalSwitchFixture(*args, **kwargs))
+        return fix.obj
+
+    def test_ls_dns_set_clear_records(self):
+        dns1 = self._dns_add()
+        dns2 = self._dns_add()
+
+        ls1 = self._ls_add('ls1')
+        self.api.ls_set_dns_records(ls1.uuid, [dns1.uuid, dns2.uuid]).execute()
+        self.assertItemsEqual([dns1.uuid, dns2.uuid],
+                              [dns.uuid for dns in ls1.dns_records])
+
+        self.api.ls_clear_dns_records(ls1.uuid).execute()
+        self.assertEqual([], ls1.dns_records)
+
+    def test_ls_dns_add_remove_records(self):
+        dns1 = self._dns_add()
+        dns2 = self._dns_add()
+
+        ls1 = self._ls_add('ls1')
+
+        self.api.ls_add_dns_record(ls1.uuid, dns1.uuid).execute()
+        self.assertItemsEqual([dns1.uuid],
+                              [dns.uuid for dns in ls1.dns_records])
+
+        self.api.ls_add_dns_record(ls1.uuid, dns2.uuid).execute()
+        self.assertItemsEqual([dns1.uuid, dns2.uuid],
+                              [dns.uuid for dns in ls1.dns_records])
+
+        self.api.ls_remove_dns_record(ls1.uuid, dns2.uuid).execute()
+        self.assertItemsEqual([dns1.uuid],
+                              [dns.uuid for dns in ls1.dns_records])
+        self.api.ls_remove_dns_record(ls1.uuid, dns1.uuid).execute()
+        self.assertEqual([], ls1.dns_records)
+
+
+class TestPortGroup(OvnNorthboundTest):
+
+    def setUp(self):
+        super(TestPortGroup, self).setUp()
+        self.switch = self.useFixture(fixtures.LogicalSwitchFixture()).obj
+        self.pg_name = 'testpg-%s' % ovsdb_utils.generate_uuid()
+
+    def test_port_group(self):
+        # Assert the Port Group was added
+        self.api.pg_add(self.pg_name).execute(check_error=True)
+        row = self.api.db_find(
+            'Port_Group',
+            ('name', '=', self.pg_name)).execute(check_error=True)
+        self.assertIsNotNone(row)
+        self.assertEqual(self.pg_name, row[0]['name'])
+        self.assertEqual([], row[0]['ports'])
+        self.assertEqual([], row[0]['acls'])
+
+        # Assert the Port Group was deleted
+        self.api.pg_del(self.pg_name).execute(check_error=True)
+        row = self.api.db_find(
+            'Port_Group',
+            ('name', '=', self.pg_name)).execute(check_error=True)
+        self.assertEqual([], row)
+
+    def test_port_group_ports(self):
+        lsp_add_cmd = self.api.lsp_add(self.switch.uuid, 'testport')
+        with self.api.transaction(check_error=True) as txn:
+            txn.add(lsp_add_cmd)
+            txn.add(self.api.pg_add(self.pg_name))
+
+        port_uuid = lsp_add_cmd.result.uuid
+
+        # Lets add the port using the UUID instead of a `Command` to
+        # exercise the API
+        self.api.pg_add_ports(self.pg_name, port_uuid).execute(
+            check_error=True)
+        row = self.api.db_find(
+            'Port_Group',
+            ('name', '=', self.pg_name)).execute(check_error=True)
+        self.assertIsNotNone(row)
+        self.assertEqual(self.pg_name, row[0]['name'])
+        # Assert the port was added from the Port Group
+        self.assertEqual([port_uuid], row[0]['ports'])
+
+        # Delete the Port from the Port Group
+        with self.api.transaction(check_error=True) as txn:
+            txn.add(self.api.pg_del_ports(self.pg_name, port_uuid))
+
+        row = self.api.db_find(
+            'Port_Group',
+            ('name', '=', self.pg_name)).execute(check_error=True)
+        self.assertIsNotNone(row)
+        self.assertEqual(self.pg_name, row[0]['name'])
+        # Assert the port was removed from the Port Group
+        self.assertEqual([], row[0]['ports'])
+
+    def test_pg_del_ports_if_exists(self):
+        self.api.pg_add(self.pg_name).execute(check_error=True)
+        non_existent_res = ovsdb_utils.generate_uuid()
+
+        # Assert that if if_exists is False (default) it will raise an error
+        self.assertRaises(RuntimeError, self.api.pg_del_ports(self.pg_name,
+                          non_existent_res).execute, True)
+
+        # Assert that if if_exists is True it won't raise an error
+        self.api.pg_del_ports(self.pg_name, non_existent_res,
+                              if_exists=True).execute(check_error=True)
